/*
 * TableSimulator.hpp
 *
 *  Created on: Jan 23, 2015
 *      Author: Joost Huizinga
 */

#ifndef SIM_SIMULATOR_HPP_
#define SIM_SIMULATOR_HPP_

#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <Sim_TouchSensor.hpp>

#include "GlutDemoApplication.h"
#include "LinearMath/btAlignedObjectArray.h"
#include "btBulletDynamicsCommon.h"
#include "GLDebugDrawer.h"


using namespace std;

class btBroadphaseInterface;
class btCollisionShape;
class btOverlappingPairCache;
class btCollisionDispatcher;
class btConstraintSolver;
struct btCollisionAlgorithmCreateFunc;
class btDefaultCollisionConfiguration;


/**
 * Callback function to process collisions with sensors.
 */
bool myContactProcessedCallback(btManifoldPoint& cp, void* body0, void* body1)
{
    TouchSensor* sensors[2];
    btCollisionObject* o1 = static_cast<btCollisionObject*>(body0);
    btCollisionObject* o2 = static_cast<btCollisionObject*>(body1);
    sensors[0] = static_cast<TouchSensor*>(o1->getUserPointer());
    sensors[1] = static_cast<TouchSensor*>(o2->getUserPointer());

    for(size_t i=0; i<2;++i){
        if(sensors[i]){
            sensors[i]->setValue(true);
        }
    }
    return false;
}

/**
 * Simulator for evolving robots.
 */
template<typename RobotType>
class Simulator : public GlutDemoApplication
{
    btAlignedObjectArray<RobotType*> m_tables;
    //keep the collision shapes, for deletion/cleanup
    btAlignedObjectArray<btCollisionShape*> m_collisionShapes;
    btAlignedObjectArray<btHingeConstraint*> m_hinges;
    btAlignedObjectArray<btRigidBody*> m_body;
    btBroadphaseInterface*  m_broadphase;
    btCollisionDispatcher*  m_dispatcher;
    btConstraintSolver* m_solver;
    btDefaultCollisionConfiguration* m_collisionConfiguration;
    bool m_pause;
    bool m_oneStep;
    int m_internalTimer;
    int m_robotUpdateSpeed;

public:

    /**
     * Constructs the simulator and sets some values that should not
     * be reset in initPhysics.
     *
     * Note: does not properly construct the simulator yet,
     * call initPhysic before trying to use the simulator.
     */
    Simulator(): GlutDemoApplication(){
        m_robotUpdateSpeed = 10;
    }

public:
    typedef RobotType robot_t;

    /**
     * Initialized the world.
     *
     * Note that constructing the TableSimulator is not enough to create a world.
     */
    void initPhysics(){
        //Unique for the evolutionary robotics class
        m_internalTimer = 0;
        m_pause = true;

        //Information for the callback function
//        tableSimulator = this;
        gContactProcessedCallback = myContactProcessedCallback;

        //Stuff for visualization
        setTexturing(true);
        setShadows(true);
        setCameraDistance(btScalar(8.));

        // Setup the basic world
        btVector3 worldAabbMin(-10000,-10000,-10000);
        btVector3 worldAabbMax(10000,10000,10000);


        m_solver = new btSequentialImpulseConstraintSolver;
        m_collisionConfiguration = new btDefaultCollisionConfiguration();
        m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
        m_broadphase = new btAxisSweep3 (worldAabbMin, worldAabbMax);
        m_dynamicsWorld = new btDiscreteDynamicsWorld(m_dispatcher,m_broadphase,m_solver,m_collisionConfiguration);
        //m_dynamicsWorld->getDispatchInfo().m_useConvexConservativeDistanceUtil = true;
        //m_dynamicsWorld->getDispatchInfo().m_convexConservativeDistanceThreshold = 0.01f;



        // Setup a big ground box
        createGround();

        //Reset scene
        clientResetScene();

        //Disable deactivation
        gDisableDeactivation = true;
    }

    /**
     * Destroy the world and perform cleanup.
     */
    void exitPhysics(){
        int i;

        for (i=0;i<m_tables.size();i++)
        {
            delete m_tables[i];
        }

        //cleanup in the reverse order of creation/initialization

        //remove the rigidbodies from the dynamics world and delete them

        for (i=m_dynamicsWorld->getNumCollisionObjects()-1; i>=0 ;i--)
        {
            btCollisionObject* obj = m_dynamicsWorld->getCollisionObjectArray()[i];
            btRigidBody* body = btRigidBody::upcast(obj);
            if (body && body->getMotionState())
            {
                delete body->getMotionState();
            }
            m_dynamicsWorld->removeCollisionObject( obj );
            delete obj;
        }

        //delete collision shapes
        for (int j=0;j<m_collisionShapes.size();j++)
        {
            btCollisionShape* shape = m_collisionShapes[j];
            delete shape;
        }

        //delete dynamics world
        delete m_dynamicsWorld;

        //delete solver
        delete m_solver;

        //delete broadphase
        delete m_broadphase;

        //delete dispatcher
        delete m_dispatcher;

        delete m_collisionConfiguration;
    }

    /**
     * Destructor, simply calls exit Physics.
     */
    virtual ~Simulator()
    {
        exitPhysics();
    }

    /**
     * Calls step for all tables in the simulation, then steps the simulation.
     */
    void step(){
        btScalar timeStep = 0.01;
        m_internalTimer++;

        if(m_internalTimer >= m_robotUpdateSpeed){
            m_internalTimer = 0;

            for (size_t i=0;i<m_tables.size();i++)
            {
                m_tables[i]->step(timeStep*m_robotUpdateSpeed);
            }
        }

        for (size_t i=0;i<m_tables.size();i++)
        {
            m_tables[i]->executeMotorCommand();
        }

        m_dynamicsWorld->stepSimulation(timeStep, 1, timeStep);
//        std::cout << m_tables[0]->getPosition().getX() << " " << m_tables[0]->getPosition().getY() << " " << m_tables[0]->getPosition().getZ() << std::endl;
    }

    /**
     * Calls step and debugDraw if either m_pause is false or oneStep is true.
     */
    virtual void clientMoveAndDisplay(){
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        if (!m_pause || (m_pause && m_oneStep))
        {
            step();
            m_dynamicsWorld->debugDrawWorld();
            m_oneStep = false;
        }

        renderme();

        glFlush();

        glutSwapBuffers();
    }

    /**
     * Performs all display operations, including clearing the depth buffers and color buffers.
     */
    virtual void displayCallback(){
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        renderme();

        //optional but useful: debug drawing
        if (m_dynamicsWorld)
            m_dynamicsWorld->debugDrawWorld();

        glFlush();
        glutSwapBuffers();
    }

    /**
     * Keyboard callback function.
     *
     * You can press 'p' for pause and 'y' to take one step if the simulation is paused.
     */
    virtual void keyboardCallback(unsigned char key, int x, int y){
        switch (key)
        {
        case 'p':
            {
                    m_pause = !m_pause;
                    break;
            }
        case 'y':  //used y because o was also doing something else and was changing the camera angle
            {
                    m_oneStep = true;
                    break;
            }
        default:
            DemoApplication::keyboardCallback(key, x, y);
        }
    }

    /**
     * Create an instance of the table simulator.
     */
    static DemoApplication* Create()
    {
        Simulator* demo = new Simulator();
        demo->myinit();
        demo->initPhysics();
        return demo;
    }


    /**
     * Creates a ground-box for collision detection.
     */
    void createGround()
    {

        // Setup a big ground box
        {
            btCollisionShape* groundShape = new btBoxShape(btVector3(btScalar(200.),btScalar(10.),btScalar(200.)));
            m_collisionShapes.push_back(groundShape);
            btTransform groundTransform;
            groundTransform.setIdentity();
            groundTransform.setOrigin(btVector3(0,-10,0));
            localCreateRigidBody(btScalar(0.),groundTransform,groundShape);
        }

    }


    /**
     * Draws the scene and two spheres for orientation.
     */
    virtual void renderme()
    {
        extern GLDebugDrawer gDebugDrawer;
        GlutDemoApplication::renderme();
        gDebugDrawer.drawSphere(btVector3(0., 0., 0.), 0.1, btVector3(0., 0., 0.));
        gDebugDrawer.drawSphere(btVector3(0., 0., 1.), 0.1, btVector3(0., 0., 1.));
        //gDebugDrawer.drawSphere(btVector3(0, 0, 0), 1.9, btVector3(1, 0, 0)); //Creates bigger sphere under robot
    }


    /**
     * Builds a robot, adds it to the world, and returns a pointer to the constructed robot.
     */
    RobotType* buildRobot(btVector3 position = btVector3(0,0,0)){
        RobotType* newRobot = new RobotType();
        newRobot->build(m_dynamicsWorld, position);
        m_tables.push_back(newRobot);
        return newRobot;
    }

    /**
     * Builds a copy of the supplied robot, adds it to the world, and returns a pointer to the constructed robot.
     *
     * Note: does not alter the supplied robot; use the return value or a call to getRobot,
     * to get a pointer to the constructed robot.
     */
    RobotType* buildRobot(RobotType* robot, btVector3 position = btVector3(0,0,0)){
        return buildRobot(*robot);
    }

    /**
     * Builds a copy of the supplied robot, adds it to the world, and returns a pointer to the constructed robot.
     *
     * Note: does not alter the supplied robot; use the return value or a call to getRobot,
     * to get a pointer to the constructed robot.
     */
    RobotType* buildRobot(RobotType& robot, btVector3 position = btVector3(0,0,0)){
        RobotType* newRobot = new RobotType(robot);
        newRobot->build(m_dynamicsWorld, position);
        m_tables.push_back(newRobot);
        return newRobot;
    }

    /**
     * Returns the indicated robot.
     */
    RobotType* getRobot(size_t index = 0){
        return m_tables[index];
    }

    /**
     * Deletes all robots.
     */
    void deleteRobots(){
        for (size_t i=0;i<m_tables.size();i++)
        {
            delete m_tables[i];
        }
        m_tables.clear();
    }

    /**
     * Disables all sensors.
     */
    void disableSensors(){
        gContactProcessedCallback = 0;
    }

    /**
     * Resets the simulator
     */
    void reset(){
        deleteRobots();
        m_internalTimer = 0;
    }

    /**
     * Sets the robot update speed.
     */
    void setRobotUpdateSpeed(int speed){
        m_robotUpdateSpeed = speed;
    }
};


#endif /* SIM_SIMULATOR_HPP_ */
